/* TroiaPreview
 * version 0.3 date 2004-09-03
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 *
 * Parameters:
 *   "Log":          generate Logfile
 *   "Res640x480":   use display-resolution 640 x 480
 *   "Res800x600":   use display-resolution 800 x 600
 *   "Res1024x768":  use display-resolution 1024 x 768
 *   "NoFullScreen": do not use fullscreen-mode
 *   "NoMouse":      do not use mouse
 *   "InvertMouse":  invert vertical mouse-axis
 */

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <gl\gl.h>
#include <gl\glu.h>

#include "Tools.h"
#include "Vector.h"
#include "Font.h"
#include "Images.h"
#include "Log.h"
#include "Textures.h"
#include "Console.h"
#include "Globals.h"
#include "Frame.h"
#include "Animations.h"
#include "Life.h"
#include "Worms.h"
#include "Ticker.h"
#include "Tracking.h"
#include "Commands.h"

//global variables
HINSTANCE AppHandle = NULL; //handle of the application
bool FullScreen = true, MouseModeOn = true;
long InvertMouse = 1; //invert vertical mouse-axis? (1 or -1)
HWND WindowHandle; //handle of window
HDC WindowDeviceHandle; //handle of devie-context of window
HGLRC OpenGlContextHandle; //handle of OpenGl-context
long ScreenX; //display-resolution
long ScreenY;
#define ScreenY_X ((float)ScreenY / (float)ScreenX)

//show picture
static void Show( )
{
	//local variables
	long X, Y, I;
	float fX, fY, fDX, fDY;

	//clear picture and depth-information
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	//settings for background and console
	glMatrixMode( GL_PROJECTION ); //set projection
	glLoadIdentity( );
	glMatrixMode( GL_MODELVIEW ); //set view
	glLoadIdentity( );
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL ); //draw filled polygons
	glEnable( GL_TEXTURE_2D ); //textures on

	//no antialiasing
	glDisable( GL_POINT_SMOOTH );
	glDisable( GL_LINE_SMOOTH );
	glDisable( GL_POLYGON_SMOOTH );
	glDisable( GL_BLEND );

	//draw background
	if( Back )
	{
		glColor3f( 1.0f, 1.0f, 1.0f );
		fDX = 2.0f / (float)BackTexCntX;
		fDY = 2.0f / (float)BackTexCntY;
		for( Y = 0; Y < BackTexCntY; Y++ )
		for( X = 0; X < BackTexCntX; X++ )
		{
			I = Y * BackTexCntX + X;
			fX = -1.0f + X * fDX;
			fY = 1.0f - Y * fDY;
			glBindTexture( GL_TEXTURE_2D, BackTexFirstNo + I );
			glBegin( GL_QUADS );
			glNormal3f( 0.0f, 0.0f, 1.0f );
			glTexCoord2f( 0.0f, 0.0f );
			glVertex3f( fX, fY, 1.0f );
			glTexCoord2f( 1.0f, 0.0f );
			glVertex3f( fX + fDX, fY, 1.0f );
			glTexCoord2f( 1.0f, 1.0f );
			glVertex3f( fX + fDX, fY - fDY, 1.0f );
			glTexCoord2f( 0.0f, 1.0f );
			glVertex3f( fX, fY - fDY, 1.0f );
			glEnd( );
		}
	}

	//draw console
	if( Console )
	{
		fX = 512.0f / (float)ScreenX;
		fY = 512.0f / (float)ScreenY;
		glColor3f( 1.0f, 1.0f, 1.0f );
		glBindTexture( GL_TEXTURE_2D, ConsoleTexNo );
		glBegin( GL_QUADS );
		glNormal3f( 0.0f, 0.0f, 1.0f );
		glTexCoord2f( 0.0f, 0.0f );
		glVertex3f( -1.0f, -1.0f, -1.0f );
		glTexCoord2f( 1.0f, 0.0f );
		glVertex3f( -1.0f + fX, -1.0f, -1.0f );
		glTexCoord2f( 1.0f, 1.0f );
		glVertex3f( -1.0f + fX, -1.0f + fY, -1.0f );
		glTexCoord2f( 0.0f, 1.0f );
		glVertex3f( -1.0f, -1.0f + fY, -1.0f );
		glEnd( );
	}

	//settings for objects in space
	glMatrixMode( GL_PROJECTION ); //set projection
	glLoadIdentity( );
	glFrustum( -0.01, 0.01, -0.01 * ScreenY_X, 0.01 * ScreenY_X, 0.01, 20.0 );
	glMatrixMode( GL_MODELVIEW ); //set view
	glLoadIdentity( );
	gluLookAt( Pos.X, Pos.Y, Pos.Z, Pos.X + Front.X, Pos.Y + Front.Y, Pos.Z + Front.Z, 0.0, 0.0, 1.0 );

	//enable antialiasing ?
	if( AntiAlias )
	{
		glEnable( GL_POINT_SMOOTH );
		glEnable( GL_LINE_SMOOTH );
		glEnable( GL_POLYGON_SMOOTH );
		glEnable( GL_BLEND );
		glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
		//glHint( GL_POLYGON_SMOOTH_HINT, GL_NICEST );
   	}

	//bind pixel texture
	glBindTexture( GL_TEXTURE_2D, PixTexNo );
	//enable texture
	glEnable( GL_TEXTURE_2D );
	//set polygon mode
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
	//show pixels
	ShowPixels( );

	//show picture
	SwapBuffers( WindowDeviceHandle );
}

//formula for interpolation
//A=f(-1), B=f(0), C=f(1), D=f(2), f(X) wanted
#define Move_f(A,B,C,D) (-A/6.0 + B/2.0 - C/2.0 + D/6.0) * XXX + (A/2.0 - B + C/2.0) * XX + (-A/3.0 - B/2 + C - D/6.0)*X + B

//move cam
static void MoveCam( )
{
	//local variables
	double CurrentTime, CurPt, X, XX, XXX;
	long CurPtInt;
	tVector P1, P2, P3, P4;

	//cam is not moving
	if( !CamMoving )
		return;

	//get time
	CurrentTime = TimeGet( );

	//one cam-point
	if( CamptsCnt == 1 )
	{
		//move camera to that cam-point
		Pos = pCampts[0].Pos;
		Front = pCampts[0].Front;
	}
	//more cam-points
	else if( CamptsCnt > 1 )
	{
		//get current cam-point-number
		CurPt = DoubleModulo( CamptSpeed * (CurrentTime - CamptRefTime), CamptsCnt );
		//get points for interpolation
		CurPtInt = (long)CurPt;
		X = CurPt - (double)CurPtInt;
		XX = X * X;
		XXX = XX * X;
		//get points for pos-interpolation
		P1 = pCampts[(CurPtInt + CamptsCnt - 1) % CamptsCnt].Pos;
		P2 = pCampts[CurPtInt].Pos;
		P3 = pCampts[(CurPtInt + 1) % CamptsCnt].Pos;
		P4 = pCampts[(CurPtInt + 2) % CamptsCnt].Pos;
		//pos-interpolation
		//P1=f(-1), P2=f(0), P3=f(1), P4=f(2), f(X) wanted
		Pos.X = Move_f( P1.X, P2.X, P3.X, P4.X );
		Pos.Y = Move_f( P1.Y, P2.Y, P3.Y, P4.Y );
		Pos.Z = Move_f( P1.Z, P2.Z, P3.Z, P4.Z );
		//get points for front-interpolation
		P1 = pCampts[(CurPtInt + CamptsCnt - 1) % CamptsCnt].Front;
		P2 = pCampts[CurPtInt].Front;
		P3 = pCampts[(CurPtInt + 1) % CamptsCnt].Front;
		P4 = pCampts[(CurPtInt + 2) % CamptsCnt].Front;
		//front-interpolation
		//P1=f(-1), P2=f(0), P3=f(1), P4=f(2), f(X) wanted
		Front.X = Move_f( P1.X, P2.X, P3.X, P4.X );
		Front.Y = Move_f( P1.Y, P2.Y, P3.Y, P4.Y );
		Front.Z = Move_f( P1.Z, P2.Z, P3.Z, P4.Z );
	}

	//repair position and front direction
	RepairPosFront( );
}

//mouse-steering
static void Mouse( )
{
	//local variables
	POINT MousePos;
	long MouseHor, MouseVer;
	bool MouseLeft, MouseRight;
	tVector Dir, Top, Left;

	//get direction over ground
	Dir = Front;
	Dir.Z = 0;
	Normalize( &Dir );
	//get top and left vectors
	SetVector( &Top, 0.0, 0.0, 1.0 );
	Left = Dir;
	Rotate( &Left, Top, -0.5 * Pi );

	//get mouse-movement
	GetCursorPos( &MousePos );
	MouseHor = MousePos.x - ScreenX / 2;
	MouseVer = MousePos.y - ScreenY / 2;
	//get mouse-buttons
	MouseLeft = (GetAsyncKeyState( VK_LBUTTON ) & 0x8000) != 0;
	MouseRight = (GetAsyncKeyState( VK_RBUTTON ) & 0x8000) != 0;

	if( MouseLeft )
	{
		if( !MouseRight )
		{
			//move forwards / backwards
			LinearCombine( &Pos, 1.0f, Pos, - MouseVer * MouseMoveStep, Dir );
			//move left / right
			LinearCombine( &Pos, 1.0f, Pos, - MouseHor * MouseMoveStep, Left );
		}
	}
	else
	{
		if( MouseRight )
		{
			//move up / down
			LinearCombine( &Pos, 1.0f, Pos, MouseVer * MouseMoveStep, Top );
		}
		else
		{
			//turn left / right
			Rotate( &Front, Top, MouseHor * MouseTurnStep );
			//look up / down
			Rotate( &Front, Left, - MouseVer * InvertMouse * MouseTurnStep );
		}
	}

	//repair position and front direction
	RepairPosFront( );

	//center mousepointer
	SetCursorPos( ScreenX / 2, ScreenY / 2 );
}

//window procedure
LRESULT CALLBACK WindowProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	//local variables
	char Line[256];
	tVector Dir, Top, Left;

	switch( uMsg )
	{
	//key was pressed
	case WM_KEYDOWN:
		//get direction over ground
		Dir = Front;
		Dir.Z = 0;
		Normalize( &Dir );
		//get top and left vectors
		SetVector( &Top, 0.0, 0.0, 1.0 );
		Left = Dir;
		Rotate( &Left, Top, -0.5 * Pi );

		//distinguish between keys
		switch( wParam )
		{
		//escape
		case VK_ESCAPE:
			//toggle console
			Console = !Console;
			//display message
			if( Console )
				ConsoleInsertLine( "[escape] => console on" );
			else
				ConsoleInsertLine( "[escape] => console off" );
			break;
		//enter
		case VK_RETURN:
			if( ! Console ) //console off
			{
				//go to start position
				SetVector( &Pos, CenterX, CenterY, CenterZ );
				SetVector( &Front, 0.0, 1.0, 0.0 );
			}
			break;
		//space
		case VK_SPACE:	
			if( ! Console ) //console off
				//restore horizontal view
				Front.Z = 0.0;
			break;
		//left arrow
		case VK_LEFT:
			if( Console ) //console on
				//move console left
				ConsoleMoveLeft( );
			else //console off
				//turn left
				Rotate( &Front, Top, - TurnStep );
			break;
		//right arrow
		case VK_RIGHT:
			if( Console ) //console on
				//move console right
				ConsoleMoveRight( );
			else //console off
				//turn right
				Rotate( &Front, Top, TurnStep );
			break;
		//up arrow
		case VK_UP:
			if( Console ) //console on
				//move console up
				ConsoleMoveUp( );
			else //console off
				//move forwards
				LinearCombine( &Pos, 1.0f, Pos, MoveStep, Dir );
			break;
		//down arrow
		case VK_DOWN:
			if( Console ) //console on
				//move console down
				ConsoleMoveDown( );
			else //console off
				//move backwards
				LinearCombine( &Pos, 1.0f, Pos, - MoveStep, Dir );
			break;
		//home
		case VK_HOME:
			if( ! Console ) //console off
				//look up
				Rotate( &Front, Left, TurnStep );
			break;
		//end
		case VK_END:
			if( ! Console ) //console off
				//look down
				Rotate( &Front, Left, - TurnStep );
			break;
		//page up
		case 0x21: //VK_PAGE_UP
			if( Console ) //console on
				//move console up a page
				ConsoleMovePageUp( );
			else //console off
				//move up
				LinearCombine( &Pos, 1.0f, Pos, MoveStep, Top );
			break;
		//page down
		case 0x22: //VK_PAGE_DOWN
			if( Console ) //console on
				//move console down a page
				ConsoleMovePageDown( );
			else //console off
				//move down
				LinearCombine( &Pos, 1.0f, Pos, - MoveStep, Top );
			break;
		//insert
		case VK_INSERT:
			if( !Console ) //console off
				//move left
				LinearCombine( &Pos, 1.0f, Pos, MoveStep, Left );
			break;
		//delete
		case VK_DELETE:
			if( !Console ) //console off
				//move right
				LinearCombine( &Pos, 1.0f, Pos, - MoveStep, Left );
			break;
		}

		//repair position and front direction
		RepairPosFront( );
		break;

	//character was typed
	case WM_CHAR:
		//console is inactive ===> do nothing
		if( !Console )
			break;
		//normal characters
		if( (unsigned char)wParam >= ' ' )
			ConsolePutc( (char)wParam );
		//enter
		if( wParam == '\r' )
		{
			//get last line of console
			GetConsoleLine( Line, sizeof( Line ) );
			//display enter
			ConsolePutc( '\n' );
			//process Command
			ProcessCommand( Line );
		}
		//backspace
		if( wParam == '\b' )
			ConsolePutc( '\b' );
		break;

	//window-close-click
	case WM_CLOSE:
		//end
		PostQuitMessage( 0 );
		break;
	}
	
	return DefWindowProc( hWnd, uMsg, wParam, lParam );
}

//main program
int WINAPI WinMain( HINSTANCE AppHndl, HINSTANCE UnusedHndl, LPSTR Parameters, int ShowMode )
{
	//local variables
	long I, J, X, Y;
	DEVMODE DevMode;
	WNDCLASS WindowClass;
	int PixelFormat;
	PIXELFORMATDESCRIPTOR PixelFormatDesc = 
	{
		sizeof( PIXELFORMATDESCRIPTOR ), //size of structure
		1, //version
		PFD_DRAW_TO_WINDOW | //supports Window
		PFD_SUPPORT_OPENGL | //supports OpenGl
		PFD_DOUBLEBUFFER, //doublebuffered
		PFD_TYPE_RGBA, //RGBA-type
		16, //16 bit color-depth
		0, 0, 0, 0, 0, 0, //ignore color-bits
		0, //no alpha-buffer
		0, //no shift-bit
		0, //no accumulation-buffer
		0, 0, 0, 0, //ignore accumulation-bits
		16, //16 bit Z-buffer
		0, //no stencil-buffer
		0, //no auxiluary buffer
		PFD_MAIN_PLANE, //main plane
		0, //reserved
		0, 0, 0 //ignore layer-masks
	};
	char Text[256], * pChr;
	MSG Message;

	//global initialisations
	AppHandle = AppHndl;
	srand( GetTickCount( ) );

	//open logfile
	if( stristr( Parameters, "Log" ) != NULL )
		LogStart( "TroiaPreview.log" );

	//resolution - autodetect
	ScreenX = GetSystemMetrics( SM_CXSCREEN );
	ScreenY = GetSystemMetrics( SM_CYSCREEN );
	//resolution - parameters
	if( (pChr = stristr( Parameters, "Res" )) != NULL )
	{
		I = 640;
		J = 480;
		if( sscanf( pChr+3, "%dx%d", &X, &Y ) == 2 )
		{
			ScreenX = LongMinMax( X, 640, 6400 );
			ScreenY = LongMinMax( Y, 400, 4000 );
		}
	}

	//mouse-mode on ?
	MouseModeOn = stristr( Parameters, "NoMouse" ) == NULL;
	//invert mouse ?
	if( stristr( Parameters, "InvertMouse" ) != NULL )
		InvertMouse = -1;

	//full-screen-mode on
	FullScreen = stristr( Parameters, "NoFullScreen" ) == NULL;
	if( FullScreen )
	{
		DevMode.dmSize = sizeof( DevMode );
		DevMode.dmBitsPerPel = 16;
		DevMode.dmPelsWidth = ScreenX;
		DevMode.dmPelsHeight = ScreenY;
		DevMode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
		ChangeDisplaySettings( &DevMode, CDS_FULLSCREEN );
	}

	//register window-class
	WindowClass.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	WindowClass.lpfnWndProc = &WindowProc; 
	WindowClass.cbClsExtra = 0;
	WindowClass.cbWndExtra = 0;
	WindowClass.hInstance = AppHandle;
	WindowClass.hIcon = NULL;
	WindowClass.hCursor = NULL;
	WindowClass.hbrBackground = (HBRUSH)COLOR_BACKGROUND;
	WindowClass.lpszMenuName = NULL;
	WindowClass.lpszClassName = "TroiaPreview";
	RegisterClass( &WindowClass );

	//create window
	WindowHandle = CreateWindow( "TroiaPreview", "TroiaPreview", WS_POPUP | WS_VISIBLE, 0, 0, ScreenX, ScreenY, NULL, NULL, AppHandle, NULL );
	while( ShowCursor( false ) >= 0 );
	SetCursor( NULL );

	//initialize OpenGl
	WindowDeviceHandle = GetDC( WindowHandle );
	PixelFormat = ChoosePixelFormat( WindowDeviceHandle, &PixelFormatDesc );
	SetPixelFormat( WindowDeviceHandle, PixelFormat, &PixelFormatDesc );
	DescribePixelFormat( WindowDeviceHandle, PixelFormat, sizeof( PixelFormatDesc ), &PixelFormatDesc );
	OpenGlContextHandle = wglCreateContext( WindowDeviceHandle );
	wglMakeCurrent( WindowDeviceHandle, OpenGlContextHandle );

	//OpenGl settings
	glDepthFunc( GL_LEQUAL );
	glEnable( GL_DEPTH_TEST );
	glEnable( GL_AUTO_NORMAL );
	glEnable( GL_NORMALIZE );
	glPixelStorei( GL_UNPACK_ALIGNMENT, 4 );
	glViewport( 0, 0, ScreenX, ScreenY );
	glLineWidth( 1.5 ); //set line width
	glPointSize( 2.0 ); //set point size

	//initialize cam-points
	CamptsCnt = 0; //no cam-points
	pCampts = (tCampt *)malloc( 0 );
	CamptSpeed = 0.0; //cam-point-speed
	CamptRefTime = 0.0; //reference-time for cam-point-movement
	CamMoving = false; //if cam is not moving

	//start-position
	SetVector( &Pos, CenterX, CenterY, CenterZ );
	SetVector( &Front, 0.0, 1.0, 0.0 );

	//initialise console
	ConsoleInit( );
	ConsoleInsertLine( "OpenGL information" );
	strcpy( Text, "  GL_VENDOR: " );
	strncat( Text, (char *)glGetString( GL_VENDOR ), sizeof( Text ) - strlen( Text ) - 1 );
	ConsoleInsertLine( Text );
	strcpy( Text, "  GL_RENDERER: " );
	strncat( Text, (char *)glGetString( GL_RENDERER ), sizeof( Text ) - strlen( Text ) - 1 );
	ConsoleInsertLine( Text );
	strcpy( Text, "  GL_VERSION: " );
	strncat( Text, (char *)glGetString( GL_VERSION ), sizeof( Text ) - strlen( Text ) - 1 );
	ConsoleInsertLine( Text );
	ConsoleInsertLine( "TroiaPreview" );
	ConsoleInsertLine( "  version 0.3 date 2004-09-03" );
	ConsoleInsertLine( "  Copyright (C) 2003-2004 1stein <1stein@schuermans.info>" );
	ConsoleInsertLine( "  Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html" );
	ConsoleInsertLine( "  some important keys:" );
	ConsoleInsertLine( "    alt-f4: exit program" );
	ConsoleInsertLine( "    escape: toggle console" );
	ConsoleInsertLine( "  when console is off:" );
	ConsoleInsertLine( "    arrow-keys: move around" );
	ConsoleInsertLine( "  when console is on:" );
	ConsoleInsertLine( "    arrow-keys: scroll console" );
	ConsoleInsertLine( "  type \"help\" for commands" );

	//initialize script buffer
	ScriptBufferSet( "scriptload TroiaPreview.ini" );

	//set time to 0
	TimeSet( 0.0 );

	//initialize frame with random data
	FrameRandomize( 0x00, 0xFF );

	//initialize
	TickerInit( );

	//center mousepointer
	if( MouseModeOn )
		SetCursorPos( ScreenX / 2, ScreenY / 2 );

	//message-loop
	Message.message = 0;

	while( Message.message != WM_QUIT )
	{
		//calls tasks working on static frame
		AnimTask( );
		LifeTask( );
		WormTask( );

		//copy static frame into temporary frame
		FrameCopyTmp( );

		//call tasks working on temporary frame
		TickerTask( );
		TrackingTask( );

		//process all messages
		while( PeekMessage( &Message, NULL, 0, 0, PM_REMOVE ) )
		{
			TranslateMessage( &Message );
			DispatchMessage( &Message );
			if( Message.message == WM_QUIT )
				break;
		}

		//mouse-steering
		if( MouseModeOn )
			Mouse( );

		//do autmatic cam movement
		MoveCam( );

		//refresh console
		if( Console )
			ConsoleRefresh( );

		//show new picture
		Show( );

		//process a few script-commands
		ProcessScriptBufferCommands( );

	} //while( Message.message != WM_QUIT )

	//unload cam-points
	free( pCampts );

	//cleanup script buffer
	ScriptBufferCleanup( );

	//cleanup OpenGl
	wglMakeCurrent( NULL, NULL );
	wglDeleteContext( OpenGlContextHandle );
	ReleaseDC( WindowHandle, WindowDeviceHandle );

	//destroy window and windows-class
	ShowCursor( true );
	DestroyWindow( WindowHandle );
	UnregisterClass( "TroiaPreview", AppHandle );

	//full-screen-mode-off
	if( FullScreen )
		ChangeDisplaySettings( NULL, 0 );

	//close logfile
	LogStop( );

	//end
    return 0;
}
